"""Test the `fio_banka` module.

Notes:
    * There does not seem to be an easy way to test GPC format encoding
        (cp1250). Therefore, the encoding is not tested.
    * Official API docs don't limit year or date values. We don't want
        to speculate, therefore they are not tested. Sanity checks only.
    * Tests of invalid formats (account statement, transaction report) are
        covered by:
        `TestAccount.test_fetch_transaction_report_for_period_with_invalid_format`,
        `TestAccount.test_fetch_account_statement_with_invalid_format`.
    * Tests of `AccountInfo` and `Transaction` objects are covered by `Account`
        tests.
"""

import datetime
import decimal
import os
import pathlib

import pytest
import requests

# Avoid shadowing by pytest fixture of the same name
import requests_mock as _requests_mock

import fio_banka


def test_constants_are_present():
    assert hasattr(fio_banka, "REQUEST_TIMELIMIT_IN_SECONDS")


def test_transaction_report_formats():
    """Test that all supported transaction report formats are present."""
    assert sorted([member.value for member in fio_banka.TransactionReportFmt]) == sorted(
        ["csv", "gpc", "html", "json", "ofx", "xml"]
    )


def test_account_statement_formats():
    """Test that all supported transaction report formats are present."""
    assert sorted([member.value for member in fio_banka.AccountStatementFmt]) == sorted(
        ["csv", "gpc", "html", "json", "ofx", "xml", "pdf", "mt940", "cba_xml", "sba_xml"]
    )


def test_exceptions_are_present():
    assert sorted([e for e in dir(fio_banka) if e.endswith("Error")]) == sorted(
        [
            "FioBankaError",
            "RequestError",
            "InvalidRequestError",
            "TimeLimitError",
            "InvalidTokenError",
            "TooManyItemsError",
            "AuthorizationError",
        ]
    )


class TestAccount:
    BASE_URL = "https://fioapi.fio.cz/v1/rest"
    TOKEN = "testTokenXZVZPOJ4pMrdnPleaUcdUlqy2LqFFVqI4dagXgi1eB1cgLzNjwsWS36"

    @pytest.fixture()
    def account(self) -> fio_banka.Account:
        return fio_banka.Account(self.TOKEN)

    @staticmethod
    @pytest.fixture()
    def account_statement() -> str:
        this_dir = pathlib.Path(os.path.realpath(__file__)).parent
        with (this_dir / "account_statement.json").open("r") as f:
            return f.read()

    @staticmethod
    @pytest.mark.parametrize(
        "token",
        [TOKEN, "", TOKEN[:-1], TOKEN + "x"],
        ids=["valid", "empty", "too-short", "too-long"],
    )
    def test_token_length(token: str):
        """Test that a token with invalid length is refused."""
        if len(token) == 64:  # noqa: PLR2004
            fio_banka.Account(token)
        else:
            with pytest.raises(ValueError, match="Invalid token length"):
                fio_banka.Account(token)

    @pytest.mark.parametrize(
        "fmt", [fio_banka.TransactionReportFmt.JSON, fio_banka.TransactionReportFmt.XML]
    )
    def test_fetch_transaction_report_for_period(
        self,
        account: fio_banka.Account,
        requests_mock: _requests_mock.Mocker,
        fmt: fio_banka.TransactionReportFmt,
    ):
        """Test the method `fetch_transaction_report_for_period`.

        The method should call correct URL and return text data.
        """
        text = "dummy text"
        requests_mock.get(
            self.BASE_URL + f"/periods/{self.TOKEN}/2023-01-01/2023-01-02/transactions.{fmt}",
            text=text,
        )
        assert (
            account.fetch_transaction_report_for_period(
                datetime.date(2023, 1, 1), datetime.date(2023, 1, 2), fmt
            )
            == text
        )

    @staticmethod
    def test_fetch_transaction_report_for_period_with_invalid_date(
        account: fio_banka.Account, requests_mock: _requests_mock.Mocker
    ):
        """Test that an invalid date range is refused."""
        requests_mock.get(_requests_mock.ANY)
        with pytest.raises(ValueError, match="Invalid date"):
            account.fetch_transaction_report_for_period(
                datetime.date(2023, 1, 2),
                datetime.date(2023, 1, 1),
                fio_banka.TransactionReportFmt.JSON,
            )

    @staticmethod
    def test_fetch_transaction_report_for_period_with_invalid_format(
        account: fio_banka.Account, requests_mock: _requests_mock.Mocker
    ):
        """Test that an invalid or unsupported data format is refused."""
        requests_mock.get(_requests_mock.ANY)
        with pytest.raises(TypeError, match="Invalid type"):
            account.fetch_transaction_report_for_period(
                datetime.date(2023, 1, 1),
                datetime.date(2023, 1, 1),
                "foo",  # type: ignore[reportArgumentType,arg-type]
            )

    @pytest.mark.parametrize(
        "fmt",
        [
            fio_banka.AccountStatementFmt.JSON,
            fio_banka.AccountStatementFmt.GPC,
            fio_banka.AccountStatementFmt.PDF,
        ],
    )
    def test_fetch_account_statement(
        self,
        account: fio_banka.Account,
        requests_mock: _requests_mock.Mocker,
        fmt: fio_banka.AccountStatementFmt,
    ):
        """Test the method `fetch_account_statement`.

        The method should call correct URL and return text or binary data.
        """
        url = self.BASE_URL + f"/by-id/{self.TOKEN}/2023/1/transactions.{fmt}"
        payload: str | bytes
        if fmt == fio_banka.AccountStatementFmt.PDF:
            payload = b"dummy bytes"
            requests_mock.get(url, content=payload)
        else:
            payload = "dummy text"
            requests_mock.get(url, text=payload)
        assert account.fetch_account_statement(2023, 1, fmt) == payload

    @staticmethod
    def test_fetch_account_statement_with_invalid_id(
        account: fio_banka.Account,
        requests_mock: _requests_mock.Mocker,
    ):
        """Test that negative ID is refused."""
        requests_mock.get(_requests_mock.ANY)
        with pytest.raises(ValueError, match="Invalid statement ID"):
            account.fetch_account_statement(2023, -1, fio_banka.AccountStatementFmt.JSON)

    @staticmethod
    def test_fetch_account_statement_with_invalid_format(
        account: fio_banka.Account, requests_mock: _requests_mock.Mocker
    ):
        """Test that an invalid or unsupported data format is refused."""
        requests_mock.get(_requests_mock.ANY)
        with pytest.raises(TypeError, match="Invalid type"):
            account.fetch_account_statement(2023, 1, "bar")  # type: ignore[reportArgumentType,arg-type]

    @pytest.mark.parametrize(
        "fmt", [fio_banka.TransactionReportFmt.HTML, fio_banka.TransactionReportFmt.OFX]
    )
    def test_fetch_transaction_report_since_last_download(
        self,
        account: fio_banka.Account,
        requests_mock: _requests_mock.Mocker,
        fmt: fio_banka.TransactionReportFmt,
    ):
        """Test the method `fetch_transaction_report_since_last_download`.

        The method should call correct URL and return text data.
        """
        text = "dummy text"
        requests_mock.get(self.BASE_URL + f"/last/{self.TOKEN}/transactions.{fmt}", text=text)
        assert account.fetch_transaction_report_since_last_download(fmt) == text

    def test_fetch_last_account_statement_info(
        self,
        account: fio_banka.Account,
        requests_mock: _requests_mock.Mocker,
    ):
        """Test the method `fetch_last_account_statement_info`.

        The method should call correct URL and return a tuple with year and ID
        of the last account statement.
        """
        requests_mock.get(self.BASE_URL + f"/lastStatement/{self.TOKEN}/statement", text="2023,12")
        assert account.fetch_last_account_statement_info() == (2023, 12)

    def test_set_last_downloaded_transaction_id(
        self,
        account: fio_banka.Account,
        requests_mock: _requests_mock.Mocker,
    ):
        """Test the method `set_last_downloaded_transaction_id`.

        It should only call correct URL. No return value is expected.
        """
        transaction_id = 10000000000
        requests_mock.get(self.BASE_URL + f"/set-last-id/{self.TOKEN}/{transaction_id}/")
        account.set_last_downloaded_transaction_id(transaction_id)

    def test_set_last_downloaded_transaction_id_with_invalid_id(
        self,
        account: fio_banka.Account,
        requests_mock: _requests_mock.Mocker,
    ):
        """Test that an invalid ID is refused."""
        requests_mock.get(_requests_mock.ANY)
        with pytest.raises(ValueError, match="Invalid transaction ID"):
            account.set_last_downloaded_transaction_id(-1)

    def test_set_last_unsuccessful_download_date(
        self,
        account: fio_banka.Account,
        requests_mock: _requests_mock.Mocker,
    ):
        """Test the method `set_last_unsuccessful_download_date`.

        It should only call correct URL. No return value is expected.
        """
        requests_mock.get(self.BASE_URL + f"/set-last-date/{self.TOKEN}/2023-12-31/")
        account.set_last_unsuccessful_download_date(datetime.date(2023, 12, 31))

    @staticmethod
    def test_parse_account_info(account_statement: str):
        """Test the method `parse_account_info`.

        It should parse a JSON string and return an object representing account
        information.
        """
        assert fio_banka.Account.parse_account_info(account_statement) == fio_banka.AccountInfo(
            account_id="2000000000",
            bank_id="2010",
            currency="CZK",
            iban="CZ1000000000002000000000",
            bic="FIOBCZPPXXX",
            opening_balance=decimal.Decimal("4000.99"),
            closing_balance=decimal.Decimal("1000.10"),
            date_start=datetime.date(2023, 1, 1),
            date_end=datetime.date(2023, 1, 3),
            year_list=None,
            id_list=None,
            id_from=10000000000,
            id_to=10000000002,
            id_last_download=None,
        )

    @staticmethod
    def test_parse_transactions(account_statement: str):
        """Test the method `parse_transactions`.

        It should parse a JSON string and return generator yielding objects
        representing individual transactions.
        """
        txns = list(fio_banka.Account.parse_transactions(account_statement))
        total_txns = 3
        assert total_txns == len(txns) == len({txn.transaction_id for txn in txns})
        assert txns[0] == fio_banka.Transaction(
            transaction_id="10000000000",
            date=datetime.date(2023, 1, 1),
            amount=decimal.Decimal("-2000.00"),
            currency="CZK",
            account_id=None,
            account_name="",
            bank_id=None,
            bank_name=None,
            ks=None,
            vs="1000",
            ss=None,
            user_identification="Nákup: example.com, dne 31.12.2022, částka  2000.00 CZK",
            remittance_info="Nákup: example.com, dne 31.12.2022, částka  2000.00 CZK",
            type="Platba kartou",
            executor="Novák, Jan",
            specification=None,
            comment="Nákup: example.com, dne 31.12.2022, částka  2000.00 CZK",
            bic=None,
            order_id=30000000000,
            payer_reference=None,
        )

    @staticmethod
    def test_request_timeout(account: fio_banka.Account, requests_mock: _requests_mock.Mocker):
        """Test that a proper exception is raised on a request timeout."""
        requests_mock.get(_requests_mock.ANY, exc=requests.exceptions.ConnectTimeout)
        with pytest.raises(fio_banka.InvalidRequestError):
            account.fetch_transaction_report_for_period(
                datetime.date(2023, 1, 1),
                datetime.date(2023, 1, 1),
                fio_banka.TransactionReportFmt.JSON,
            )

    @staticmethod
    @pytest.mark.parametrize(
        ("status_code", "exception"),
        [
            (404, fio_banka.InvalidRequestError),
            (409, fio_banka.TimeLimitError),
            (413, fio_banka.TooManyItemsError),
            (422, fio_banka.AuthorizationError),
            (500, fio_banka.InvalidTokenError),
            (599, fio_banka.RequestError),
        ],
    )
    def test_request_error_status_codes(
        account: fio_banka.Account,
        requests_mock: _requests_mock.Mocker,
        status_code: int,
        exception: type[fio_banka.FioBankaError],
    ):
        """Test that a proper exception is raised in case of a request error."""
        requests_mock.get(_requests_mock.ANY, status_code=status_code)
        with pytest.raises(exception):
            account.fetch_transaction_report_for_period(
                datetime.date(2023, 1, 1),
                datetime.date(2023, 1, 1),
                fio_banka.TransactionReportFmt.JSON,
            )



"""Fio banka API.

API description:
    https://www.fio.cz/docs/cz/API_Bankovnictvi.pdf (Czech only)

`REQUEST_TIMELIMIT_IN_SECONDS`: time limit for 1 request.
`AccountStatementFmt`: enum of account statement formats.
`TransactionReportFmt`: enum of transaction report formats.
`AccountInfo`: named tuple of account information.
`Transaction`: named tuple of a single transaction data.
`FioBankaError`: base exception.
    `RequestError`: an HTTP request could not be fulfilled.
        `InvalidRequestError`: request is invalid (typically URL or token).
        `TimeLimitError`: request time limit has been exceeded.
        `InvalidTokenError`: token is inactive or invalid.
        `TooManyItemsError`: the number of transactions exceeds 50000.
        `AuthorizationError`: token is not authorized to fetch data (e.g. data
            older then N days).
`Account`: client for interaction with an account.
    `fetch_transaction_report_for_period()`: return transaction report for
        a given time period.
    `fetch_account_statement()`: return account statement.
    `fetch_transaction_report_since_last_download()`: return transaction report
        since the last download.
    `fetch_last_account_statement_info()`: return year and ID of the last
        account statement.
    `set_last_downloaded_transaction_id()`: set ID of the last successfully
        downloaded transaction.
    `set_last_unsuccessful_download_date()`: set date of the last unsucessful
        download.
    `parse_account_info()`: return account information from a JSON string.
    `parse_transactions()`: return transactions from a JSON string.

Example usage:
    >>> import fio_banka, datetime
    >>> account = fio_banka.Account("my-API-token")
    >>> transaction_report = account.fetch_transaction_report_for_period(
    ...     datetime.date(2023, 1, 1),
    ...     datetime.date(2023, 1, 2),
    ...     fio_banka.TransactionReportFmt.JSON
    ... )
    >>> account.parse_account_info(transaction_report)  # JSON only
    AccountInfo(
        account_id='2000000000',
        bank_id='2010',
        currency='CZK',
        iban='CZ1000000000002000000000',
        ...
    )
    >>> next(iter(account.parse_transactions(transaction_report)))  # JSON only
    Transaction(
        transaction_id='10000000000',
        date=datetime.date(2023, 1, 1),
        amount=Decimal('2000.0'),
        currency='CZK',
        account_id=None,
        ...
"""

import collections.abc
import datetime
import decimal
import enum
import json
import typing

import requests

REQUEST_TIMELIMIT_IN_SECONDS = 30


@enum.unique
class AccountStatementFmt(enum.StrEnum):
    """Enum of account statement formats."""

    CSV = enum.auto()
    GPC = enum.auto()
    HTML = enum.auto()
    JSON = enum.auto()
    OFX = enum.auto()
    XML = enum.auto()
    PDF = enum.auto()
    MT940 = enum.auto()
    CBA_XML = enum.auto()
    SBA_XML = enum.auto()


@enum.unique
class TransactionReportFmt(enum.StrEnum):
    """Enum of transaction report formats."""

    CSV = enum.auto()
    GPC = enum.auto()
    HTML = enum.auto()
    JSON = enum.auto()
    OFX = enum.auto()
    XML = enum.auto()


_OptionalStr = str | None
_OptionalDecimal = decimal.Decimal | None
_OptionalDate = datetime.date | None
_OptionalInt = int | None
_Fmt = AccountStatementFmt | TransactionReportFmt


class AccountInfo(typing.NamedTuple):
    """Account information."""

    account_id: _OptionalStr
    bank_id: _OptionalStr
    currency: _OptionalStr
    iban: _OptionalStr
    bic: _OptionalStr
    opening_balance: _OptionalDecimal
    closing_balance: _OptionalDecimal
    date_start: _OptionalDate
    date_end: _OptionalDate
    year_list: _OptionalInt
    id_list: _OptionalInt
    id_from: _OptionalInt
    id_to: _OptionalInt
    id_last_download: _OptionalInt


class Transaction(typing.NamedTuple):
    """Transaction data."""

    transaction_id: str
    date: datetime.date
    amount: decimal.Decimal
    currency: str
    account_id: _OptionalStr
    account_name: _OptionalStr
    bank_id: _OptionalStr
    bank_name: _OptionalStr
    ks: _OptionalStr
    vs: _OptionalStr
    ss: _OptionalStr
    user_identification: _OptionalStr
    remittance_info: _OptionalStr
    type: _OptionalStr
    executor: _OptionalStr
    specification: _OptionalStr
    comment: _OptionalStr
    bic: _OptionalStr
    order_id: _OptionalInt
    payer_reference: _OptionalStr


class FioBankaError(Exception):
    """Base exception."""


class RequestError(FioBankaError):
    """An HTTP request could not be fulfilled."""

    def __init__(self) -> None:  # noqa: D107
        super().__init__(self.__class__.__doc__)


class InvalidRequestError(RequestError):
    """Request is invalid (typically URL or token)."""


class TimeLimitError(RequestError):
    """Request time limit has been exceeded."""


class InvalidTokenError(RequestError):
    """Token is inactive or invalid."""


class TooManyItemsError(RequestError):
    """The number of transactions exceeds 50000."""


class AuthorizationError(RequestError):
    """Token is not authorized to fetch data (e.g. data older than N days)."""


def _parse_date(fio_date: str) -> datetime.date:
    return datetime.date.fromisoformat(fio_date[:10])


def _parse_json(data: str):
    return json.loads(data, parse_float=decimal.Decimal)


class Account:
    """Client for interaction with an account."""

    _BASE_URL = "https://fioapi.fio.cz/v1/rest"
    _TOKEN_LEN = 64
    # https://requests.readthedocs.io/en/latest/user/advanced/#timeouts
    # It's a good practice to set connect timeouts to slightly larger than
    # a multiple of 3, which is the default TCP packet retransmission window.
    _TIMEOUT_IN_SECONDS = 10

    def __init__(self, token: str) -> None:
        """Return Account instance.

        Args:
            token (str): API token (64 characters long)
        """
        if len(token) != self._TOKEN_LEN:
            raise ValueError(f"Invalid token length: token must have {self._TOKEN_LEN} characters")
        self._token = token

    def _request(self, url: str, fmt: _Fmt | None) -> requests.Response:
        # IMPORTANT: As the API token is included in the URL, make sure it
        # is not shown in traceback. Use the `from` clause with caution.
        try:
            response = requests.get(self._BASE_URL + url, timeout=self._TIMEOUT_IN_SECONDS)
        except requests.exceptions.RequestException:
            # Timeout is typically hit when using an invalid token.
            raise InvalidRequestError from None
        try:
            response.raise_for_status()
        except requests.HTTPError:
            exception: type[RequestError]
            match response.status_code:
                case 404:
                    exception = InvalidRequestError
                case 409:
                    exception = TimeLimitError
                case 413:
                    exception = TooManyItemsError
                case 422:
                    exception = AuthorizationError
                case 500:
                    exception = InvalidTokenError
                case _:
                    exception = RequestError
            raise exception from None
        match fmt:
            case TransactionReportFmt.GPC | AccountStatementFmt.GPC:
                response.encoding = "cp1250"
        return response

    @staticmethod
    def _check_fmt(fmt: _Fmt, fmt_type: type[_Fmt]) -> None:
        if not isinstance(fmt, fmt_type):
            raise TypeError(f"Invalid type: `fmt` must be a member of {fmt_type.__name__} enum")

    def _request_transaction_report(self, url: str, fmt: TransactionReportFmt) -> str:
        self._check_fmt(fmt, TransactionReportFmt)
        return self._request(url, fmt).text

    def _request_account_statement(self, url: str, fmt: AccountStatementFmt) -> str | bytes:
        self._check_fmt(fmt, AccountStatementFmt)
        response = self._request(url, fmt)
        if fmt == AccountStatementFmt.PDF:
            return response.content  # bytes
        return response.text

    def fetch_transaction_report_for_period(
        self, date_from: datetime.date, date_to: datetime.date, fmt: TransactionReportFmt
    ) -> str:
        """Return transaction report for the given time period.

        Args:
            date_from (datetime.date): start date
            date_to (datetime.date): end date
            fmt (TransactionsReportFmt): transaction report format

        Raises:
            RequestError or any of its subclasses.

        Side effects:
            Calls Fio banka API.

        Returns:
            str: transaction report
        """
        if date_from > date_to:
            raise ValueError("Invalid date: `date_from` must be less than or equal to `date_to`")
        url = (
            f"/periods/{self._token}/{date_from.isoformat()}/{date_to.isoformat()}/"
            f"transactions.{fmt}"
        )
        return self._request_transaction_report(url, fmt)

    def fetch_account_statement(
        self, year: int, statement_id: int, fmt: AccountStatementFmt
    ) -> str | bytes:
        """Return account statement.

        Args:
            year (int): year of the account statement
            statement_id (int): ID of the account statement
            fmt (AccountStatementFmt): account statement format

        Raises:
            RequestError or any of its subclasses.

        Side effects:
            Calls Fio banka API.

        Returns:
            str | bytes: account statement (bytes when the format is PDF, str
                otherwise)
        """
        if statement_id < 0:
            raise ValueError(f"Invalid statement ID: '{statement_id}'; ID must be positive")
        url = f"/by-id/{self._token}/{year}/{statement_id}/transactions.{fmt}"
        return self._request_account_statement(url, fmt)

    def fetch_transaction_report_since_last_download(self, fmt: TransactionReportFmt) -> str:
        """Return transaction report since the last download.

        Args:
            fmt (TransactionReportFmt): transaction report format

        Raises:
            RequestError or any of its subclasses.

        Side effects:
            Calls Fio banka API.

        Returns:
            str: transaction report
        """
        url = f"/last/{self._token}/transactions.{fmt}"
        return self._request_transaction_report(url, fmt)

    def fetch_last_account_statement_info(self):
        """Return year and ID of the last account statement.

        Raises:
            RequestError or any of its subclasses.

        Side effects:
            Calls Fio banka API.

        Returns:
            tuple[int, int]: account statement year and ID
        """
        url = f"/lastStatement/{self._token}/statement"
        year, statement_id = self._request(url, None).text.split(",")
        return (int(year), int(statement_id))

    def set_last_downloaded_transaction_id(self, transaction_id: int) -> None:
        """Set ID of the last successfully downloaded transaction.

        Args:
            transaction_id (int): transaction ID

        Raises:
            RequestError or any of its subclasses.

        Side effects:
            Calls Fio banka API.
        """
        if transaction_id < 0:
            raise ValueError(f"Invalid transaction ID: '{transaction_id}'; ID must be positive")
        url = f"/set-last-id/{self._token}/{transaction_id}/"
        self._request(url, None)

    def set_last_unsuccessful_download_date(self, download_date: datetime.date) -> None:
        """Set date of the last unsuccessful download.

        Args:
            download_date (datetime.date): date of the last unsuccessful download

        Raises:
            RequestError or any of its subclasses.

        Side effects:
            Calls Fio banka API.
        """
        url = f"/set-last-date/{self._token}/{download_date.isoformat()}/"
        self._request(url, None)

    @staticmethod
    def parse_account_info(data: str) -> AccountInfo:
        """Return account information from data.

        Args:
            data (str): a JSON string representing transaction report or
                account statement

        Returns:
            AccountInfo: account information
        """
        info = _parse_json(data)["accountStatement"]["info"]
        return AccountInfo(
            account_id=info["accountId"],
            bank_id=info["bankId"],
            currency=info["currency"],
            iban=info["iban"],
            bic=info["bic"],
            opening_balance=info["openingBalance"],
            closing_balance=info["closingBalance"],
            date_start=_parse_date(info["dateStart"]),
            date_end=_parse_date(info["dateEnd"]),
            year_list=info["yearList"],
            id_list=info["idList"],
            id_from=info["idFrom"],
            id_to=info["idTo"],
            id_last_download=info["idLastDownload"],
        )

    @staticmethod
    def parse_transactions(data: str) -> collections.abc.Generator[Transaction, None, None]:
        """Yield transactions from data.

        Args:
            data (str): a JSON string representing transaction report or
                account statement

        Yields:
            Generator[Transaction, None, None]: transaction
        """

        def get_value(data, key, cast: collections.abc.Callable | None = None):
            value = data[key]["value"]
            if cast is not None:
                return cast(value)
            return value

        def get_value_or_none(data, key):
            if data[key] is None:
                return None
            return get_value(data, key)

        txns = _parse_json(data)["accountStatement"]["transactionList"]["transaction"]
        for txn in txns:
            yield Transaction(
                transaction_id=get_value(txn, "column22", cast=str),  # str
                date=get_value(txn, "column0", cast=_parse_date),
                amount=get_value(txn, "column1"),
                currency=get_value(txn, "column14"),
                account_id=get_value_or_none(txn, "column2"),
                account_name=get_value_or_none(txn, "column10"),
                bank_id=get_value_or_none(txn, "column3"),
                bank_name=get_value_or_none(txn, "column12"),
                ks=get_value_or_none(txn, "column4"),
                vs=get_value_or_none(txn, "column5"),
                ss=get_value_or_none(txn, "column6"),
                user_identification=get_value_or_none(txn, "column7"),
                remittance_info=get_value_or_none(txn, "column16"),
                type=get_value_or_none(txn, "column8"),
                executor=get_value_or_none(txn, "column9"),
                specification=get_value_or_none(txn, "column18"),
                comment=get_value_or_none(txn, "column25"),
                bic=get_value_or_none(txn, "column26"),
                order_id=get_value_or_none(txn, "column17"),  # int
                payer_reference=get_value_or_none(txn, "column27"),
            )
